from django.contrib.auth.models import User
from django.template import Template, Context
from django.test import TestCase, RequestFactory


class TemplateTagsTest(TestCase):
    def test_templatetags_loaded(self):
        request = RequestFactory().get('/')
        request.user = User.objects.create(username='test')
        Template('{% load speedbar %}{% metric "overall" "time" %}').render(Context({'request': request}))




from django.http import HttpResponse
from django.core.cache import cache
from django.contrib.admin.views.decorators import staff_member_required
from speedbar.utils import DETAILS_PREFIX, TRACE_PREFIX

import json

@staff_member_required
def panel(request, trace_id):
    details = cache.get(DETAILS_PREFIX + trace_id)
    if details:
        details_json = json.dumps(details, skipkeys=True, default=repr, indent=2) # Cannot use decorator as need default=repr
        return HttpResponse(content=details_json, content_type='text/javascript; charset=utf-8')
    return HttpResponse(status=404)

@staff_member_required
def trace(request, trace_id):
    trace = cache.get(TRACE_PREFIX + trace_id)
    if trace:
        return HttpResponse(json.dumps(trace))
    return HttpResponse(status=404)


def noop():
    pass



from __future__ import absolute_import
from .base import BaseModule, RequestTrace
from .stacktracer import trace_method

try:
    from redis import StrictRedis
except ImportError:
    StrictRedis = None

class RedisModule(BaseModule):
    key = 'redis'

    def get_metrics(self):
        return RequestTrace.instance().stacktracer.get_node_metrics('REDIS')

    def get_details(self):
        redis_nodes = RequestTrace.instance().stacktracer.get_nodes('REDIS')
        return [{'operation': node.extra['operation'], 'key': node.extra['key'], 'time': node.duration} for node in redis_nodes]

def init():
    if StrictRedis is None:
        return False

    # The linter thinks the methods we monkeypatch are not used
    # pylint: disable=W0612
    @trace_method(StrictRedis)
    def execute_command(self, *args, **kwargs):
        if len(args) >= 2:
            action = 'Redis: %s (%s)' % args[:2]
            key = args[1]
        else:
            action = 'Redis: %s' % args[:1]
            key = ''
        return ('REDIS', action, {'operation': args[0], 'key': key})

    return RedisModule



from __future__ import absolute_import

from django.template import defaulttags
from django.template.base import add_to_builtins, Library, Template
from django.template.response import TemplateResponse
from django.template.loader_tags import BlockNode

from .stacktracer import trace_method, trace_function

register = Library()


class DecoratingParserProxy(object):
    """
    Mocks out the django template parser, passing templatetags through but
    first wrapping them to include performance data
    """
    def __init__(self, parser):
        self.parser = parser

    def add_library(self, library):
        wrapped_library = Library()
        wrapped_library.filters = library.filters
        for name, tag_compiler in library.tags.items():
            wrapped_library.tags[name] = self.wrap_compile_function(name, tag_compiler)
        self.parser.add_library(wrapped_library)

    def wrap_compile_function(self, name, tag_compiler):
        def compile(*args, **kwargs):
            node = tag_compiler(*args, **kwargs)
            node.render = trace_function(node.render, ('TEMPLATE_TAG', 'Render tag: ' + name, {}))
            return node
        return compile


@register.tag
def load(parser, token):
    decorating_parser = DecoratingParserProxy(parser)
    return defaulttags.load(decorating_parser, token)


# The linter thinks the methods we monkeypatch are not used
# pylint: disable=W0612
def init():
    # Our eventual aim is to patch the render() method on the Node objects
    # corresponding to custom template tags. However we have to jump through
    # a number of hoops in order to get access to the object.
    #
    # 1. Add ourselves to the set of built in template tags
    #    This allows us to replace the 'load' template tag which controls
    #    the loading of custom template tags
    # 2. Delegate to default load with replaced parser
    #    We provide our own parser class so we can catch and intercept
    #    calls to add_library.
    # 3. add_library receives a library of template tags
    #    It iterates through each template tag, wrapping its compile function
    # 4. compile is called as part of compiling the template
    #    Our wrapper is called instead of the original templatetag compile
    #    function. We delegate to the original function, but then modify
    #    the resulting object by wrapping its render() function. This
    #    render() function is what ends up being timed and appearing in the
    #    tree.
    add_to_builtins('speedbar.modules.templates')

    @trace_method(Template)
    def __init__(self, *args, **kwargs):
        name = args[2] if len(args) >= 3 else '<Unknown Template>'
        return ('TEMPLATE_COMPILE', 'Compile template: ' + name, {})

    @trace_method(Template)
    def render(self, *args, **kwargs):
        name = self.name if self.name is not None else '<Unknown Template>'
        return ('TEMPLATE_RENDER', 'Render template: ' + name, {})

    @trace_method(BlockNode)
    def render(self, *args, **kwargs):
        return ('BLOCK_RENDER', 'Render block: ' + self.name, {})

    @trace_method(TemplateResponse)
    def resolve_context(self, *args, **kwargs):
        return ('TEMPLATE_CONTEXT', 'Resolve context', {})



import socket

from .base import BaseModule


class HostInformationModule(BaseModule):
    key = 'host'

    def get_metrics(self):
        return {'name': socket.gethostname()}


def init():
    return HostInformationModule



from __future__ import absolute_import
from .base import BaseModule, RequestTrace
from .stacktracer import trace_method

try:
    import memcache
except ImportError:
    memcache = None

MEMCACHE_OPERATIONS = [ 'add', 'append', 'cas', 'decr', 'delete', 'get', 'gets', 'incr', 'prepend', 'replace', 'set', ]
MEMCACHE_MULTI_OPERATIONS = [ 'get_multi', 'set_multi', 'delete_multi', ]


class MemcacheModule(BaseModule):
    key = 'memcache'

    def get_metrics(self):
        return RequestTrace.instance().stacktracer.get_node_metrics('MEMCACHE')

    def get_details(self):
        memcache_nodes = RequestTrace.instance().stacktracer.get_nodes('MEMCACHE')
        return [{'operation': node.extra['operation'], 'key': node.extra['key'], 'time': node.duration} for node in memcache_nodes]


# The linter thinks the methods we monkeypatch are not used
# pylint: disable=W0612
def intercept_memcache_operation(operation):
    @trace_method(memcache.Client, operation)
    def info(self, *args, **kwargs):
        return ('MEMCACHE', 'Memcache: %s (%s)' % (operation, args[0]), {'operation': operation, 'key': args[0]})


def intercept_memcache_multi_operation(operation):
    @trace_method(memcache.Client, operation)
    def info(self, *args, **kwargs):
        return ('MEMCACHE', 'Memcache: %s' % (operation,), {'operation': operation, 'key': ''})


def init():
    if memcache is None:
        return False

    for operation in MEMCACHE_OPERATIONS:
        intercept_memcache_operation(operation)

    for operation in MEMCACHE_MULTI_OPERATIONS:
        intercept_memcache_multi_operation(operation)

    return MemcacheModule



from __future__ import absolute_import

from django.utils.html import format_html
from django import template
register = template.Library()

@register.simple_tag
def metric(module, metric):
    """
    Display a placeholder that the middleware converts to a particular summary metric
    """
    return format_html(u'<span data-module="{0}" data-metric="{1}"></span>', module, metric)


try:
    # Django <1.6
    from django.conf.urls.defaults import *
except ImportError:
    # Django >=1.6
    from django.conf.urls import *


def foo(request):
    from django.http import HttpResponse
    return HttpResponse()

urlpatterns = patterns('',
    url('^$', foo, name='speedbar_test_url'),
    (r'^speedbar/', include('speedbar.urls')),
)


from setuptools import setup, find_packages

tests_require = [
    'Django>=1.4,<1.9',
]

install_requires = [
    'ProxyTypes>=0.9',
]

setup(
    name='django-speedbar',
    version='0.2.2',
    author='Mat Clayton',
    author_email='mat@mixcloud.com',
    url='http://github.com/mixcloud/django-speedbar',
    description='Provides a break down of page loading time',
    long_description=open('README.rst').read(),
    packages=find_packages(exclude=["tests",]),
    install_requires=install_requires,
    license='MIT License',
    tests_require=tests_require,
    extras_require={'test': tests_require},
    test_suite="testrunner.runtests",
    include_package_data=True,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



"""
Speedbar module

This provides performance metrics, details of operations performed, and Chrome SpeedTracer integration
for page loads.

Information is provided by a set of modules, which are responsible for recording and reporting data.
The collected data is then collected and made available via template tags, headers, and a HAR file.

On startup each module is given a chance to initialize itself, typically this consists of monkey
patching a set of built in django functionality. A per request module object is created in response
to the start of each request. Over the course of the request modules record data, using thread local
storage to associate correctly with the right request. A middleware then writes out summary data,
and the headers required to fetch more detailed information from the server. Finally the request_finished
signal handler stores detailed information to memcache which can then be retrieved.

"""
import re

from django.conf import settings
from django.core.signals import request_started, request_finished
from django.core.urlresolvers import reverse
from django.utils.encoding import smart_unicode, smart_str

from speedbar.signals import setup_request_tracing, store_request_trace
from speedbar.utils import init_modules
from speedbar.modules.base import RequestTrace


if getattr(settings, 'SPEEDBAR_ENABLE', True):
    # We hook everything up in the middleware file as loading the middleware is one of the first things performed
    # by the django WSGI implementation.
    init_modules()

    request_started.connect(setup_request_tracing, dispatch_uid='request_started_speedbar_setup_request_tracing')
    request_finished.connect(store_request_trace, dispatch_uid='request_started_speedbar_store_request_trace')


HTML_TYPES = ('text/html', 'application/xhtml+xml')
METRIC_PLACEHOLDER_RE = re.compile('<span data-module="(?P<module>[^"]+)" data-metric="(?P<metric>[^"]+)"></span>')


class SpeedbarMiddleware(object):
    """
    Middleware module to add speedbar related headers to respones and replace any
    speedbar template tags with their correct values.
    For most accurate results place this as near to the top of your middleware stack
    as possible.

    If you wish to customize how features are enabled and disabled on a per-request
    basis you can do so by inheriting from this class and overriding some or all of the
    should_ methods.
    """
    def process_request(self, request):
        if getattr(settings, 'SPEEDBAR_ENABLE', True):
            request_trace = RequestTrace.instance()
            request_trace.stacktracer.root.label = '%s %s' % (request.method, request.path)
            request_trace.request = request

    def process_response(self, request, response):
        if not getattr(settings, 'SPEEDBAR_ENABLE', True):
            return response

        request_trace = RequestTrace.instance()
        # TODO: Do we also need to stash this on in case of exception?
        request_trace.response = response

        metrics = dict((key, module.get_metrics()) for key, module in request_trace.modules.items())

        if self.should_return_response_headers(request):
            self.add_response_headers(response, metrics)

        if self.should_return_trace_header(request):
            response['X-TraceUrl'] = reverse('speedbar_trace', args=[request_trace.id])
            request_trace.persist_log = True

        if self.should_replace_template_tags(request):
            if 'gzip' not in response.get('Content-Encoding', '') and response.get('Content-Type', '').split(';')[0] in HTML_TYPES:

                # Force render of response (from lazy TemplateResponses) before speedbar is injected
                if hasattr(response, 'render'):
                    response.render()
                content = smart_unicode(response.content)

                content = self.replace_templatetag_placeholders(content, metrics)

                # Note: The URLs returned here do not exist at this point. The relevant data is added to the cache by a signal handler
                # once all page processing is finally done. This means it is possible summary values displayed and the detailed
                # break down won't quite correspond.
                if getattr(settings, 'SPEEDBAR_PANEL', True):
                    panel_url = reverse('speedbar_panel', args=[request_trace.id])
                    panel_placeholder_url = reverse('speedbar_details_for_this_request')
                    content = content.replace(panel_placeholder_url, panel_url)
                    request_trace.persist_details = True

                response.content = smart_str(content)
                if response.get('Content-Length', None):
                    response['Content-Length'] = len(response.content)
        return response

    def should_return_response_headers(self, request):
        return getattr(settings, 'SPEEDBAR_RESPONSE_HEADERS', False)

    def should_return_trace_header(self, request):
        return hasattr(request, 'user') and request.user.is_staff and getattr(settings, 'SPEEDBAR_TRACE', True)

    def should_replace_template_tags(self, request):
        return hasattr(request, 'user') and request.user.is_staff

    def add_response_headers(self, response, metrics):
        """
        Adds all summary metrics to the response headers, so they can be stored in nginx logs if desired.
        """
        def sanitize(string):
            return string.title().replace(' ', '-')

        for module, module_values in metrics.items():
            for key, value in module_values.items():
                response['X-Speedbar-%s-%s' % (sanitize(module), sanitize(key))] = value

    def replace_templatetag_placeholders(self, content, metrics):
        """
        The templatetags defined in this module add placeholder values which we replace with true values here. They
        cannot just insert the values directly as not all processing may have happened by that point.
        """
        def replace_placeholder(match):
            module = match.group('module')
            metric = match.group('metric')
            return unicode(metrics[module][metric])
        return METRIC_PLACEHOLDER_RE.sub(replace_placeholder, content)
